/*-------------- Telecommunications & Signal Processing Lab ---------------
                             McGill University

Routine:
  AFILE *AFwrAUhead (FILE *fp, int Format, long int Nframe, long int Nchan,
                     double Sfreq)

Purpose:
  Write header information to an AU audio file

Description:
  This routine opens and writes header information to an AU audio file.  The
  additional header information is written as part of the file header.

  AU audio file header:
      Bytes     Type    Contents
     0 ->  3    int    File identifier
     4 ->  7    int    Header size H (bytes)
     8 -> 11    int    Audio data length (bytes)
    12 -> 15    int    Data encoding format
    16 -> 19    int    Sample rate (samples per second)
    20 -> 23    int    Number of interleaved channels
    24 -> 27    int    AFsp identifier ("AFsp)")
    28 -> H-1   --     Additional header information

Parameters:
  <-  AFILE *AFwrAUhead
      Audio file pointer for the audio file.  In case of error, the audio file
      pointer is set to NULL.
   -> FILE *fp
      File pointer for the audio file
   -> int Format
      Audio file data format code: FD_MULAW8, FD_ALAW8, FD_INT8, FD_INT16, etc.
   -> long int Nframe
      Number of samples.  This value is used to set the file header for
      non-random access files.  If this value is AF_NFRAME_UNDEF, the number
      of samples is unknown.
   -> long int Nchan
      Number of channels
   -> double Sfreq
      Sampling frequency

Author / revision:
  P. Kabal  Copyright (C) 2001
  $Revision: 1.73 $  $Date: 2001/10/04 03:01:14 $

-------------------------------------------------------------------------*/

static char rcsid [] = "$Id: AFwrAUhead.c 1.73 2001/10/04 AFsp-v6r5a $";

#include <assert.h>
#include <setjmp.h>
#include <string.h>

#include <libtsp.h>
#include <libtsp/nucleus.h>
#include <libtsp/AFheader.h>
#include <libtsp/AFmsg.h>
#define AF_DATA_LENGTHS
#include <libtsp/AFpar.h>
#include <libtsp/AUpar.h>

#define ICEILV(n,m)	(((n) + ((m) - 1)) / (m))	/* int n,m >= 0 */
#define RNDUPV(n,m)	((m) * ICEILV (n, m))		/* Round up */
#define WRPAD(fp,size,align) \
     AFwriteHead (fp, NULL, 1, (int) (RNDUPV(size, align) - (size)), \
		  DS_NATIVE);

#define MCOPY(src,dest)		memcpy ((void *) (dest), \
					(const void *) (src), sizeof (dest))

#define ALIGN		4	/* Header length is a multiple of ALIGN */

/* setjmp / longjmp environment */
extern jmp_buf AFW_JMPENV;

static double
AF_ScaleF (int Format);
static int
AF_setDencod (int Format);


AFILE *
AFwrAUhead (FILE *fp, int Format, long int Nframe, long int Nchan,
	    double Sfreq)

{
  struct AU_head Fhead;
  AFILE *AFp;
  int Lhead;
  struct AF_write AFw;

/* Set the long jump environment; on error return a NULL */
  if (setjmp (AFW_JMPENV))
    return NULL;	/* Return from a header write error */

/* Set up the encoding parameters */
  AFw = AFpreSetWPar (Format, Nchan, Sfreq);
  AFw.DFormat.ScaleF = AF_ScaleF (Format);
  AFw.DFormat.Swapb = DS_EB;

/* Leave room for the header information */
  if (AFw.Hinfo.N > 0)
    Lhead = RNDUPV (AU_LHMIN + (sizeof FM_AFSP) - 1 + AFw.Hinfo.N, ALIGN);
  else
    Lhead = RNDUPV (AU_LHMIN + 4L, ALIGN);

/* Set up the fixed header parameters */
  MCOPY (FM_AU, Fhead.Magic);
  Fhead.Lhead = Lhead;
  if (Nframe == AF_NFRAME_UNDEF)
    Fhead.Ldata = AU_NOSIZE;
  else
    Fhead.Ldata = Nframe * Nchan * AF_DL[Format];
  Fhead.Dencod = AF_setDencod (Format);
  Fhead.Srate = (uint4_t) (Sfreq + 0.5);	/* Rounding */
  Fhead.Nchan = (uint4_t) Nchan;

/* Write out the header */
  WHEAD_S (fp, Fhead.Magic);
  WHEAD_V (fp, Fhead.Lhead, DS_EB);
  WHEAD_V (fp, Fhead.Ldata, DS_EB);
  WHEAD_V (fp, Fhead.Dencod, DS_EB);
  WHEAD_V (fp, Fhead.Srate, DS_EB);
  WHEAD_V (fp, Fhead.Nchan, DS_EB);

/* AFsp information records */
  if (AFw.Hinfo.N > 0) {
    WHEAD_SN (fp, FM_AFSP, (sizeof FM_AFSP) - 1);	/* Omit null char */
    WHEAD_SN (fp, AFw.Hinfo.Info, AFw.Hinfo.N);
    WRPAD (fp, AFw.Hinfo.N, ALIGN);
  }
  else
    WHEAD_SN (fp, "\0\0\0\0", 4);

/* Set the parameters for file access */
  AFp = AFsetWrite (fp, FT_AU, &AFw);

  return AFp;
}

/* Set the scale factor */


static double
AF_ScaleF (int Format)

{
  double ScaleF;

  switch (Format) {
  case FD_MULAW8:
    ScaleF = 1./AU_SF_MULAW8;
    break;
  case FD_ALAW8:
    ScaleF = 1./AU_SF_ALAW8;
    break;
  case FD_INT8:
    ScaleF = 1./AU_SF_LIN8;
    break;
  case FD_INT16:
    ScaleF = 1./AU_SF_LIN16;
    break;
  case FD_INT24:
    ScaleF = 1./AU_SF_LIN24;
    break;
  case FD_INT32:
    ScaleF = 1./AU_SF_LIN32;
    break;
  case FD_FLOAT32:
    ScaleF = 1./AU_SF_FLOAT32;
    if (! UTcheckIEEE ())
      UTwarn ("AFwrAUhead - %s", AFM_NoIEEE);
    break;
  case FD_FLOAT64:
    ScaleF = 1./AU_SF_DOUBLE64;
    if (! UTcheckIEEE ())
      UTwarn ("AFwrAUhead - %s", AFM_NoIEEE);
    break;
  default:
    UTwarn ("AFwrAUhead - %s", AFM_AU_UnsData);
    ScaleF = 0.0;	/* Error */
    break;
  }

  return ScaleF;
}

/* Set up the encoding parameters */


static int
AF_setDencod (int Format)

{
  int Dencod;

  switch (Format) {
  case FD_MULAW8:
    Dencod = AU_MULAW8;
    break;
  case FD_ALAW8:
    Dencod = AU_ALAW8;
    break;
  case FD_INT8:
    Dencod = AU_LIN8;
    break;
  case FD_INT16:
    Dencod = AU_LIN16;
    break;
  case FD_INT24:
    Dencod = AU_LIN24;
    break;
  case FD_INT32:
    Dencod = AU_LIN32;
    break;
  case FD_FLOAT32:
    Dencod = AU_FLOAT32;
    break;
  case FD_FLOAT64:
    Dencod = AU_DOUBLE64;
    break;
  default:
    Dencod = AU_UNSPEC;  /* Error */
    assert (0);
  }

  return Dencod;
}
